#include "screen.h"

SCREEN_DATA Screen; 

void ScreenInit(int width, int height){
  // Init emulated screen:
  Screen.width=width;
  Screen.height=height;
  int px_width=width/64;
  int px_height=height/32;
  Screen.pixel=(px_width < px_height) ? px_width: px_height;
  Screen.originX=(width-64*Screen.pixel)/2;
  Screen.originY=(height-32*Screen.pixel)/2;
  ScreenClear();

  SetTraceLogLevel(LOG_ERROR); // Disable anoying raylib logs
  InitWindow(width, height, "Chip-8 Emulator");
  SetTargetFPS(80); // Set game to run at 60 frames-per-second
}

void ScreenClear() {
  for(int i=0;i<64*32;i++){
    Screen.pixels[i]=0;
  }
}

void ScreenUpdate(){
  BeginDrawing();
  ClearBackground(RAYWHITE);
  //  DrawText("Congrats! You created your first window!", 190, 200, 20, LIGHTGRAY);
  for(int x=0;x<64;x++){
    for(int y=0;y<32;y++){
      if(Screen.pixels[x+y*64] == 0)
        DrawRectangle(Screen.originX+Screen.pixel*x,Screen.originY+Screen.pixel*y,Screen.pixel,Screen.pixel,BLACK);
      else
        DrawRectangle(Screen.originX+Screen.pixel*x,Screen.originY+Screen.pixel*y,Screen.pixel,Screen.pixel,WHITE);
    }
  }  

  EndDrawing();
}

char ScreenPixelApply(int x, int y, unsigned char state){
  if(Screen.pixels[x+y*64] != 0 && state != 0)
    return 1;
  Screen.pixels[x+y*64]=state;
  return 0;
}

void ScreenPixelFlip(int x, int y){
  if(Screen.pixels[x+y*64]==0)
    Screen.pixels[x+y*64]=1;
  else
    Screen.pixels[x+y*64]=0;
}

void ScreenWH(int *width, int *height){
  *width=64;
  *height=32;
}

void ScreenClose(){
  CloseWindow(); // Close window and OpenGL context
}


